#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stdint.h>
#include <limits.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/time.h>

/* 导入统一 SQL 转换函数 */
extern
int TransferSQL(int sourceDbTypeCode, int targetDbTypeCode, char* sourceSqlStr, char* targetSqlStr, int targetSqlStrLen, char* jsonParameter);

static inline long long current_time_in_us() {
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return (long long)tv.tv_sec * 1000000 + tv.tv_usec;
}

static int sourceDbTypeCode = 11; // 11 oracle
// static int targetDbTypeCode = 16; // 16 GaussDBOracle
static int targetDbTypeCode = 2; // 2 lightdb-oracle
static char* jsonParameter = "{\"dbName\":\"usersvr\",\"sqlParameters\":\"[[1$$'2']]\"}";

/* 按照官方做法调用统一 SQL */
char *tsfm(char *sql, char **result_buffer, int *result_buffer_size) {
    // 转换结果 0表示成功，>0表示目标串长度不够，<0表示失败
    int transRet = 0;
    int new_size;

    // 调用TransferSQL转换函数
    transRet = TransferSQL(sourceDbTypeCode, targetDbTypeCode, sql, *result_buffer, *result_buffer_size, jsonParameter);
    if (transRet > 0) {
        // 预设目标SQL的长度不够，重新分配内存
        new_size = transRet + 1;
        *result_buffer=(char*) realloc (*result_buffer, new_size) ;
        *result_buffer_size = new_size;
        transRet = TransferSQL(sourceDbTypeCode, targetDbTypeCode, sql, *result_buffer, *result_buffer_size, jsonParameter);
    }
    if (transRet == 0) {
        // printf("After transfer sql is:%s\n",result_buffer);
    } else if (transRet < 0) {
        printf("SQL Convert Failed:%d\n",transRet);
        return NULL;
    }
    return *result_buffer;
}

typedef struct {
    int thread_id;
    int loop_count;
    int test_count;
    char **sqls;
    char **answers;
    char fail[4096];
} thread_data_t;

void* thread_function(void *arg) {
    thread_data_t *data = (thread_data_t*)arg;
    int thread_id = data->thread_id;
    int loop_count = data->loop_count;
    int test_count = data->test_count;
    char **answers = data->answers;
    char **sqls = data->sqls;
    int loop, i;
    int result_buffer_size = 8192;
    char *result_buffer = (char*) malloc(result_buffer_size * sizeof(char));

    for (loop = 0; loop < loop_count; ++loop) {
        for (i = 0; i < test_count; ++i) {
            char *result = tsfm(sqls[i], &result_buffer, &result_buffer_size);
            if (strcmp(result, answers[i]) != 0) {
                sprintf(data->fail, "Thread %d run %d check failed.\n  SQL: %s\n  EXP: %s\n  ACT: %s\n\n", thread_id, i, sqls[i], answers[i], result);
                free(result_buffer);
                return NULL;
            }
        }
    }
    free(result_buffer);
    return NULL;
}

typedef struct {
    int i;
    char *sql;
    char **answer_ptr;
} init_thread_func_arg;

/* caller should free the result */
char* read_file_to_string(const char* filename) {
    FILE *file = fopen(filename, "r");
    if (file == NULL) {
        perror("无法打开文件");
        return NULL;
    }

    fseek(file, 0, SEEK_END);
    long file_size = ftell(file);
    rewind(file);

    char *file_content = (char *)malloc((file_size + 1) * sizeof(char));
    if (file_content == NULL) {
        perror("内存分配失败");
        fclose(file);
        return NULL;
    }

    size_t read_size = fread(file_content, sizeof(char), file_size, file);
    if (read_size != file_size) {
        perror("读取文件内容失败");
        free(file_content);
        fclose(file);
        return NULL;
    }

    file_content[file_size] = '\0';
    fclose(file);
    return file_content;
}

bool test_sql_template_first(const char* sql_template, const int test_count)
{
    long long start_time;
    long long end_time;
    long long total_time;
    int i;
    char sql[65535];

    sprintf(sql, sql_template, 0);
    printf("测试 SQL 长度（字节）：%ld\n", strlen(sql));

    /* 初始化验证结果，首次调用相当于无缓存 */
    total_time = 0;
    for (i = 0; i < test_count; i++) {
        int result_buffer_size = 8192;
        char *result_buffer = (char*) malloc(result_buffer_size * sizeof(char));
        sprintf(sql, sql_template, i);
        start_time = current_time_in_us();
        tsfm(sql, &result_buffer, &result_buffer_size);
        end_time = current_time_in_us();
        total_time += (end_time - start_time);
        free(result_buffer);
    }
    printf("(首次无缓存) 总时间: %lldus, SQL 个数: %d, 平均执行耗时: %lfus\n", total_time, test_count, (double)total_time / test_count);
    return true;
}
bool test_sql_template(const int parallel, const char* sql_template, const int test_count, const int loop_count)
{
    long long start_time;
    long long end_time;
    long long total_time;
    long long total_count;
    char **sqls;
    char **answers;
    int i;
    char sql[65535];

    sqls = (char**) malloc(test_count * sizeof(char*));
    answers = (char**) malloc(test_count * sizeof(char*));

    total_count = test_count * loop_count;

    sprintf(sql, sql_template, 0);

    /* 初始化验证结果，首次调用相当于无缓存 */
    total_time = 0;
    for (i = 0; i < test_count; i++) {
        int result_buffer_size = 8192;
        char *result_buffer = (char*) malloc(result_buffer_size * sizeof(char));
        sprintf(sql, sql_template, i);
        sqls[i] = strdup(sql);
        start_time = current_time_in_us();
        char *result = tsfm(sql, &result_buffer, &result_buffer_size);
        end_time = current_time_in_us();
        total_time += (end_time - start_time);
        answers[i] = result;
    }

    {
        // 利用 answers 测试并行行为是否正确
        const int functional_parallel = 8;
        pthread_t threads[functional_parallel];
        thread_data_t thread_data[functional_parallel];
        bool failed = false;
        for (i = 0; i < functional_parallel; ++i) {
            thread_data[i].thread_id = i + 1;
            thread_data[i].loop_count = 1;
            thread_data[i].test_count = test_count;
            thread_data[i].sqls = sqls;
            thread_data[i].answers = answers;
            thread_data[i].fail[0] = '\0';
            int rc = pthread_create(&threads[i], NULL, thread_function, &(thread_data[i]));
            if (rc) {
                printf("Error: unable to create thread %d\n", i);
                exit(-1);
            }
        }
        for (i = 0; i < functional_parallel; ++i) {
            pthread_join(threads[i], NULL);
        }
        for (i = 0; i < functional_parallel; ++i) {
            if (thread_data[i].fail[0] != '\0') {
                printf("%s\n", thread_data[i].fail);
                failed = true;
            }
        }
        if (failed) {
            printf("并发测试 c 功能有错\n");
            return false;
        } else {
            printf("并发测试 c 功能通过\n");
        }
    }

    // 测试并行带缓存的执行速度
    {
        pthread_t threads[parallel];
        thread_data_t thread_data[parallel];
        bool failed = false;
        
        total_time = 0;
        start_time = current_time_in_us();
        
        printf("【C 缓存】开始测试，线程数：%d, 每个线程循环次数：%d, 不同的SQL个数：%d\n", parallel, loop_count, test_count);
        for (i = 0; i < parallel; ++i) {
            thread_data[i].thread_id = i + 1;
            thread_data[i].loop_count = loop_count;
            thread_data[i].test_count = test_count;
            thread_data[i].sqls = sqls;
            thread_data[i].answers = answers;
            thread_data[i].fail[0] = '\0';
            int rc = pthread_create(&threads[i], NULL, thread_function, &(thread_data[i]));
            if (rc) {
                printf("Error: unable to create thread %d\n", i);
                exit(-1);
            }
        }
        for (i = 0; i < parallel; ++i) {
            pthread_join(threads[i], NULL);
        }

        end_time = current_time_in_us();
        total_time = end_time - start_time;
        printf("并发数：%d, 总时间: %lldus, 每个线程运行次数: %lld, 用户体感耗时: %lfus, 总吞吐量：%lf/s\n",
               parallel, total_time, total_count, (double)total_time / total_count, (double)total_count * parallel / (total_time / 1000000));
        
        for (i = 0; i < parallel; ++i) {
            if (thread_data[i].fail[0] != '\0') {
                printf("%s\n", thread_data[i].fail);
                failed = true;
            }
        }
        if (failed) {
            printf("并发测试 c 有错\n\n");
            return false;
        } else {
            printf("并发测试 c 通过\n\n");
        }
    }

    for (i = 0; i < test_count; i++) {
        free(sqls[i]);
        free(answers[i]);
    }
    free(sqls);
    free(answers);

    // total_time = 0;
    // start_time = current_time_in_us();
    // for (i = 0; i < test_count; i++) {
    //     char sql[1024];
    //     sprintf(sql, sql_template, i);
    //     test_c_cache(sql, loop_count);
    // }
    // end_time = current_time_in_us();
    // total_time = end_time - start_time;
    // printf("(串行 C ) Time: %lldus, count: %lld, tps: %lfus\n", total_time, total_count, (double)total_time / total_count);
    return true;
}

// 执行命令并返回结果
void exec_command(const char* cmd, char* result, size_t result_size) {
    FILE *fp;
    if ((fp = popen(cmd, "r")) == NULL) {
        perror("执行命令失败");
        return;
    }

    // 读取命令的输出
    if (fgets(result, result_size, fp) != NULL) {
        // 移除末尾的换行符
        size_t len = strlen(result);
        if (len > 0 && result[len - 1] == '\n') {
            result[len - 1] = '\0';
        }
    }
    pclose(fp);
}

void print_cpu_info() {
    char result[1024];

    // 获取 CPU 型号
    result[0] = '\0';
    exec_command("lscpu | grep 'Model name' | awk -F ':' '{print $2}' |xargs", result, sizeof(result));
    printf("CPU 型号: %s\n", result);

    // 获取逻辑核心数量
    result[0] = '\0';
    exec_command("lscpu | grep '^CPU(s):' | awk '{print $2}'", result, sizeof(result));
    printf("CPU 逻辑核心数量: %s\n", result);

    // 先尝试获取 CPU max MHz
    result[0] = '\0';
    exec_command("lscpu | grep 'CPU max MHz:' | awk '{print $4}'", result, sizeof(result));
    
    if (strlen(result) == 0) {
        // 如果 CPU max MHz 不存在，使用 CPU MHz
        result[0] = '\0';
        exec_command("lscpu | grep 'CPU MHz:' | awk '{print $3}'", result, sizeof(result));
        printf("CPU 当前频率: %s MHz\n", result);
    } else {
        printf("CPU 最大频率: %s MHz\n", result);
    }
}

int main()
{
    char *template_str;
    const int test_count = 5000;
    const int parallel_array[] = {1, 2, 4, 8, 16, 32};
    const int parallel_array_size = sizeof(parallel_array) / sizeof(int);

    print_cpu_info();

    int i;

    {
        const int loop_count = 20000;
        template_str = read_file_to_string("simple-sql.sql");
        test_sql_template_first(template_str, test_count);
        for (i = 0; i < parallel_array_size; i++) {
            int parallel = parallel_array[i];
            test_sql_template(parallel, template_str, test_count, loop_count / parallel);
        }
        free(template_str);
    }

    {
        const int loop_count = 2000;
        template_str = read_file_to_string("complex-sql.sql");
        test_sql_template_first(template_str, test_count);
        for (i = 0; i < parallel_array_size; i++) {
            int parallel = parallel_array[i];
            test_sql_template(parallel, template_str, test_count, loop_count / parallel);
        }
        free(template_str);
    }

    return 0;
}

